#include "stdafx.h"
#include "detectobjects.h"
#include "dump.h"
#include "integralhistogram.h"
#include "profiling.h"
#include "prune_subwindows.h"

#include <sstream>
#include <time.h>

using namespace std;

pair<CvMat*, CvMat*> ClassifyImagePixels(CvMat * source_image,
                                         CvMat * cluster_means_norm,
                                         PixClass pixclass)
{
    CvMat * timg = cvCreateMat(source_image->rows, source_image->cols, CV_16SC1);
    CvMat * wimg = cvCreateMat(source_image->rows, source_image->cols, CV_32FC1);
    ilClassifyImagePixels(source_image, timg, wimg, cluster_means_norm, (int)pixclass);
    return make_pair(timg, wimg);
}


CvMat ** IntegralBlockHist(CvMat * timg, CvMat * wimg,
                           size_t cluster_number,
                           size_t dx, size_t dy)
{
    size_t nby = (div(timg->rows, dy)).quot;
    size_t nbx = (div(timg->cols, dx)).quot;
    CvMat ** iblockhist = new CvMat*[cluster_number];
    for (size_t i = 0; i < cluster_number; i++)
        iblockhist[i] = cvCreateMat(nby, nbx, CV_32FC1);

//    ilIntegralBlockHist2( timg, wimg, iblockhist, cluster_number, dx, dy);
//    Tick("ilIntegralBlockHist2");

    ilIntegralBlockHist( timg, wimg, iblockhist, cluster_number, dx, dy);
//    Tick("IntegralBlockHist");
    return iblockhist;
}

IMPEXP
SubWindows DetectObjects( IplImage* src,
                          Model const & model, 
                          CvMat * cmeans,
                          CvMat * covmat,
                          CvMat * cmeansnorm,
						  int pixclass,
						  HistBlockSizeArray bsarray,
						  ScaleFctArray scarray,
						  FlipFlagArray fliparray,
						  XYRatioArray xyratio,
						  RotAngleArray anglearray,
						  int borderext,
						  bool pruneflag,
						  int min_stage_num)
{
    using namespace std;

    double dt;
    int64 t1,t2,t3,t01,t02;
	clock_t tc1,tc2;
	double totproctime=0, totdettime=0;

	//*** convert to one channel CvMat
    CvMat* srcgr = cvCreateMat(src->height,src->width,CV_8UC1);
	if (src->nChannels==1) cvCopy(src,srcgr);
	else cvCvtColor(src,srcgr,CV_RGB2GRAY);

	//*** extend/shrink borders of an image
    CvMat* srcext = cvCreateMat(2*borderext+src->height,2*borderext+src->width,CV_8UC1);
	double extmat[] = {1.0, 0.0, (src->width)/2.0, 0.0, 1.0, (src->height)/2.0};
	CvMat *mapmatext = cvCreateMat(2,3,CV_64FC1);
	cvSetData(mapmatext,extmat,24);
	cvGetQuadrangleSubPix(srcgr,srcext,mapmatext);
	cvReleaseMat(&mapmatext);

	//cvNamedWindow("Extension",1);
	//cvShowImage("Extension",srcext);
	//cvWaitKey(0);
	
	SubWindows subwindows;

	//*** iterate over horizontal and vertical image flips
	// ff = 0 : no flipping
	//    = 1 : x-flipping
	//    = 2 : y-flipping
	//    = 3 : x-y-flipping
	for (FlipFlagArray::iterator ff = fliparray.begin(); ff != fliparray.end(); ++ff){

		CvMat* srcflip = cvCreateMat(srcext->height,srcext->width,CV_8UC1);
		cvCopy(srcext,srcflip);
		if (*ff==1) cvFlip(srcflip,NULL,1);
		if (*ff==2) cvFlip(srcflip,NULL,0);
		if (*ff==3) cvFlip(srcflip,NULL,-1);

		//*** iterate over scales
		for (ScaleFctArray::iterator sc = scarray.begin(); sc != scarray.end(); ++sc){

			//*** iterate over xy-ratios
			for (XYRatioArray::iterator xyr = xyratio.begin(); xyr != xyratio.end(); ++xyr){

				double xscf=*sc, yscf=*sc;
				if (*xyr>1.0) xscf=xscf*(*xyr);
				if (*xyr<1.0) yscf=yscf/(*xyr);
				int yszsc=floor(yscf*(srcext->height));
				int xszsc=floor(xscf*(srcext->width));	

				if ((xszsc>0) && (yszsc>0)){
					//*** subsample image
					//printf("xsz: %d, ysz: %d\n",xsz,ysz);
					CvMat* srcsc = cvCreateMat(yszsc,xszsc,CV_8UC1);
					cvResize(srcflip,srcsc);

					//*** iterate over rotations
					for (RotAngleArray::iterator ra = anglearray.begin(); ra != anglearray.end(); ++ra){

						double angle = *ra; 
						//double angle = -M_PI/12; 
						double rmat[] = {cos(angle), sin(angle), xszsc/2, -sin(angle), cos(angle), yszsc/2};
						ilRectangle r=ilRectangleRotateCenter(ilRectangle(0.0,0.0,(double)xszsc,(double)yszsc),angle,xszsc/2.0,yszsc/2.0);
						//printf("rotated rectangle %d,%d,%d,%d to %1.1f,%1.1f,%1.1f,%1.1f\n",0,0,xsz,ysz,r.x,r.y,r.w,r.h);
						int xsz=(int)(r.w);
						int ysz=(int)(r.h);
						CvMat* srcrot = cvCreateMat(ysz,xsz,CV_8UC1);
						CvMat *mapmat = cvCreateMat(2,3,CV_64FC1);
						cvSetData(mapmat,rmat,24);
						cvGetQuadrangleSubPix(srcsc,srcrot,mapmat);
						cvReleaseMat(&mapmat);

						//*** copy 
						CvMat* srcimg = cvCreateMat(ysz,xsz,CV_8UC1);
						cvCopy(srcrot,srcimg);
						CvMat *srcimg32f = cvCreateMat(ysz,xsz,CV_32FC1);
						cvConvert(srcimg,srcimg32f);

						CvMat* imgdisp1 = cvCreateMat(ysz,xsz,CV_8UC1);
						cvCopy(srcrot,imgdisp1);					
						CvMat* imgdisp2 = cvCreateMat(yszsc,xszsc,CV_8UC1);
						cvCopy(srcsc,imgdisp2);	
		
						//********************************************************
						//***  MAIN PART FROM HERE....
						//***  COMPUTES HISTOGRAM OF IMAGE "srcimg"	
	
						//*** label image pixels
						//int pixclass=TEXTONS_SC12_GDORDER2_;
						//int pixclass=ORIENTGRAD_DNUM4_;
						t01=cvGetTickCount();
						t1=cvGetTickCount();
						CvMat *timg = cvCreateMat(ysz,xsz,CV_16SC1);
						CvMat *wimg = cvCreateMat(ysz,xsz,CV_32FC1);
	
						tc1 = clock ();
						ilClassifyImagePixels(srcimg,timg,wimg,cmeansnorm,pixclass);
						if (0) { //DEBUG
							cvSet(timg, cvScalar(1.0));
							cvSet(wimg, cvScalar(1.0));
						}
						tc2 = clock ();
						totproctime+=(double)(tc2-tc1)/(double)CLK_TCK;
						//printf("\n  ClassifyImagePixels time:        %1.6fsec.\n",(double)(tc2-tc1)/(double)CLK_TCK);
						t2=cvGetTickCount();
						dt=(double(t2-t1)/cvGetTickFrequency())/pow(10.0,6);	
				
						//*** iterate over block sizes
						for (HistBlockSizeArray::iterator bs = bsarray.begin(); bs != bsarray.end(); ++bs){	

							//*** compute integral weighted block histogram
							t1=cvGetTickCount();
							double dx=bs->dx, dy=bs->dy;
							int nc;

							if ((dx<xsz) && (dy<ysz)){
								//printf("  - dx:%1.1f, dy:%1.1f\n",dx,dy);
								if (pixclass==TEXTONS_SC12_GDORDER2_) nc=12;
								if (pixclass==ORIENTGRAD_DNUM4_) nc=4;
								int	nbx=floor((double)xsz/(double)dx);
								int nby=floor((double)ysz/(double)dy);
								CvMat **iblockhist = new CvMat*[nc];
								for (int i=0; i<nc; i++) iblockhist[i]=cvCreateMat(nby,nbx,CV_32FC1);
								// /*
					
								tc1 = clock ();
								if (pixclass==ORIENTGRAD_DNUM4_){
									ilIntegralBlockHist(timg,wimg,iblockhist,nc,dx,dy); // weighted hist
								}else{
									ilIntegralBlockHist(timg,NULL,iblockhist,nc,dx,dy); // unweighted hist
								}
								tc2 = clock ();
								totproctime+=(double)(tc2-tc1)/(double)CLK_TCK;
								//printf("  IntegralBlockHist time:          %1.6fsec.\n",(double)(tc2-tc1)/(double)CLK_TCK);

								t3=cvGetTickCount();
								dt=(double(t3-t1)/cvGetTickFrequency())/pow(10.0,6);
								//printf("histogram computing time: %1.5fsec.\n",dt);	

								//***  COMPLETED
								//***  HISTOGRAM IS IN "iblockhist"
								//********************************************************
		
								//*** detect objects
								tc1 = clock ();
								double dtsum=0.0;
								uint hist_size_x = iblockhist[0]->cols;
								uint hist_size_y = iblockhist[0]->rows;
								uint subwin_size_x = model.begin()->histblock_winsize_x;
								uint subwin_size_y = model.begin()->histblock_winsize_y;
								uint subwin_step_x = 1;
								uint subwin_step_y = 1;    
								int top, left;
					
								if (hist_size_x>=subwin_size_x && hist_size_y>=subwin_size_y){
									int left_max = hist_size_x - subwin_size_x;     assert( left_max > 0 );
									int top_max  = hist_size_y - subwin_size_y;     assert( top_max > 0 );
									ClearModelCache(model);
									//*** classify each subwindow in the image
									for(left = 0; left < left_max; left += subwin_step_x){
										for(top = 0; top < top_max; top += subwin_step_y){
											t1=cvGetTickCount();
											SubWindowInfo result = adacascadeclass(model, iblockhist,
																				nc, left, top, 
																				subwin_size_x, subwin_size_y);
											//double stage_min=19;
											//if (result.stage>=stage_min){
											//if (result.passed){
											if (((min_stage_num>0) & (result.stage>=min_stage_num)) | (result.passed)){
												//printf("detection conf:%1.1f; rect:[%d,%d,%d,%d]\n",result.conf,(int)(result.x),(int)(result.y),result.w,result.h);

												//*** rotating
												ilRectangle rr=ilRectangleRotateCenter(ilRectangle(result.x,result.y,result.w,result.h),angle,(double)xsz/(dx*2.0),(double)ysz/(dy*2.0));
												//printf("detected rect:[%d,%d,%d,%d]; angle:%1.1f; cx:%f; cy:%f\n",result.x,result.y,result.w,result.h,angle,(double)xsz/2.0,(double)ysz/2.0);
												//printf("rotated rect:[%1.1f,%1.1f,%1.1f,%1.1f]\n",rr.x,rr.y,rr.w,rr.h);

												int w1x=(result.x+1)*dx, w1y=(result.y+1)*dy; int w1w=result.w*dx, w1h=result.h*dy;
												cvRectangle(imgdisp1,cvPoint(w1x,w1y),cvPoint(w1x+w1w,w1y+w1h),cvScalar(255,255,255),1);
												cvRectangle(imgdisp2,cvPoint(w1x,w1y),cvPoint(w1x+w1w,w1y+w1h),cvScalar(0,0,0),1);
												int w2x=(rr.x+1)*dx-(xsz-xszsc)*.5, w2y=(rr.y+1)*dy-(ysz-yszsc)*.5; int w2w=rr.w*dx, w2h=rr.h*dy;
												//int w2x=(rr.x+1)*dx, w2y=(rr.y+1)*dy; int w2w=w2x+rr.w*dx, w2h=w2y+rr.h*dy;
												cvRectangle(imgdisp2,cvPoint(w2x,w2y),cvPoint(w2x+w2w,w2y+w2h),cvScalar(255,255,255),1);
											
												//*** scaling and rotation adjustment
												result.x = (int)(w2x/xscf);
												result.y = (int)(w2y/yscf);
												result.w = (int)(w2w/xscf);
												result.h = (int)(w2h/yscf);
												//*** older scaling
												//result.x = (int)((double)(result.x+1)*dx/xscf);
												//result.y = (int)((double)(result.y+1)*dy/yscf);
												//result.w *= dx/xscf;
												//result.h *= dy/yscf;
												//result.conf = result.stage;
												//*** flipping
												if ((*ff==1) || (*ff==3)) result.x=(srcext->cols)-result.x-result.w;
												if ((*ff==2) || (*ff==3)) result.y=(srcext->rows)-result.y-result.h;
												//if (*ff==3) {result.x=(srcext->cols)-result.x; result.y=(srcext->rows)-result.y;}
												//** border extension
												result.x = result.x-borderext;
												result.y = result.y-borderext;

												//*** check if detection has sufficient overlap (=0.5) with the image
												double bi_x1=max(0,result.x), bi_x2=min(src->width,result.x+result.w);
												double bi_y1=max(0,result.y), bi_y2=min(src->height,result.y+result.h);
												double bi_w=bi_x2-bi_x1, bi_h=bi_y2-bi_y1;
												//printf("checking rectangle [%d,%d,%d,%d]\n",result.x,result.y,result.w,result.h);
												if (bi_w*bi_h/(result.w*result.h)>0.7){
													subwindows.push_back(result);
													//printf(" -> passed\n");
												}
											}
											t2=cvGetTickCount();
											dt=(double(t2-t1)/cvGetTickFrequency())/pow(10.0,6);
											dtsum+=dt;
										}
									} // iteration over all positions
									//printf("Detection time: %1.3fsec.\n",dtsum);
									t02=cvGetTickCount();
									dt=(double(t02-t01)/cvGetTickFrequency())/pow(10.0,6);
									//printf("Total detection time %1.3f\n",dt);
								}
								tc2 = clock ();
								totdettime+=(double)(tc2-tc1)/(double)CLK_TCK;
								//printf("  Detection time:                  %1.6fsec.\n",(double)(tc2-tc1)/(double)CLK_TCK);
			
								// */
								for (int i=0; i<nc; i++) cvReleaseMat(&iblockhist[i]);
								delete[] *iblockhist;
							}
						} // iteration over block sizes				
					
						//cvNamedWindow("Detection1",1);
						//cvShowImage("Detection1",imgdisp1);
						//cvNamedWindow("Detection2",1);
						//cvShowImage("Detection2",imgdisp2);
						//cvWaitKey(0);
						cvReleaseMat(&srcimg);
						cvReleaseMat(&srcimg32f);
						cvReleaseMat(&imgdisp1);
						cvReleaseMat(&imgdisp2);
						cvReleaseMat(&srcrot);
						cvReleaseMat(&timg);
						cvReleaseMat(&wimg);
					}// iteration over rotations
					cvReleaseMat(&srcsc);
				}
			} // iterarion over xy-ratios
		} // iterarion over scales
		cvReleaseMat(&srcflip);
	} // iterarion over flip flags
	
	//printf(" -> total processing time: %1.3f;\n",totproctime);
	//printf("    total detection time:  %1.3f\n",totdettime);

	cvReleaseMat(&srcext);
	cvReleaseMat(&srcgr);

	//*** pruning
	if (pruneflag){
		SubWindows prunedsubwindows;
	    PruneSubWindows(subwindows, 0.3, 1, prunedsubwindows);
		return prunedsubwindows;
	} else {
      return subwindows;
	}
}

void ClearModelCache(Model const & model)
{
    for (Model::const_iterator it = model.begin(),
                              end = model.end();
                              it != end; ++it)
    {
        for (StageRules::const_iterator rit = it->rules.begin(),
                                       rend = it->rules.end();
                                       rit != rend; ++rit)
        {
            rit->clear_cache();
        }
    }
}

void RotateRectCenter()
{

}