#include "stdafx.h"
#include "ilcharsegment.h"


#define CHARSEGMENT_DEBUG		0
#define CHARSEGMENT_WRITEXML	0


//****************************************************************************************
// int ilNPCharRead(const CvArr *srcarr, string *npstr, int thr0, int thrstep, int nthr)
//
//   In the given image 'srcarr' reads a line of characters (designed for car number plates).
//   To segment characters from the image, a set of gray-value thresholds are applied defined
//   as follows: thr={thr0, thr0+thrstep, ..., thr0+nthr*thrstep}. The output of each
//   segmentation passed to the character recognition module that outputs a string of chars.
//   The string that has been recognized consistently in most of the tries is
//   returned in "npstr". The confidence returned by the function is the number of
//   times the recognized string has been read (at different threshold levels).
//
//   ADVISES: For the better performance in general use more and fine threshold levels 
//   (large "nthr", small "thrstep") if this is OK for the speed. The input image
//   does not have to be a carefully segmented number plate. Being 2,3 times larger than
//   the number plate is fine if you want to make sure you deinitely got the whole number
//   in the image. Running the algo. on full images (cars, streets, etc.) is problematic
//   right now, so some step of pre-segmentation is required.
//
//   DEPENDENCIES: The code makes use of a Neuural Network character recognition implemented
//   in Matlab. To interface this library link the code to "libnncharrec.lib". Other Matlab
//   libraries will be required. On machines without matlab these can be installed using
//   MCRInstaller. If you need to recompile the shared matlab library use this command 
//   from the directory "nncharrec" in matlab: >mcc -B csharedlib:libnncharrec nncharrec.m -v
//*****************************************************************************************
int ilNPCharRead(const CvArr *srcarr, string *npstr, int thr0, int thrstep, int nthr, CvMat **charrectout){
	int thr;
	CvMat **npcharrect = new CvMat* [nthr];
	int *scount = new int[nthr];
	string *strall = new string[nthr];
	
	for (int i=0; i<nthr; i++){
		thr=thr0+thrstep*i;
		int succread=ilNPCharReadAtThresh(srcarr,&(strall[i]),&(npcharrect[i]),thr);
		scount[i]=0;
		
#if CHARSEGMENT_DEBUG
		printf("Using thresh=%3d read characters: ",thr);
		for (int j=0; j<(strall[i]).length(); j++)
			printf("%c",(strall[i])[j]);
		printf("\n");
#endif
	}

	int nmax=0, imax=-1;
	for (int i=0; i<nthr; i++){
		if ((strall[i]).length()>2){
			for (int j=0; j<nthr; j++){
				if (!strall[i].compare(strall[j])){
					scount[i]++;
					if (scount[i]>nmax){
						nmax=scount[i];
						imax=i;
					}
				}
			}
		}
	}

	(*npstr)="";
	if (imax>=0) {
		(*npstr)+=strall[imax];
		if (charrectout){
			if ((npcharrect[imax]->rows>0)&&(npcharrect[imax]->cols>0)){
				*charrectout=cvCreateMat(npcharrect[imax]->rows,npcharrect[imax]->cols,npcharrect[imax]->type);
				cvCopy(npcharrect[imax],*charrectout);
			} else {
				*charrectout=NULL;
			}
		}
	}

#if CHARSEGMENT_DEBUG
		if (imax>=0){
		printf("Read %d characters with confidence %d: ",(*npstr).length(),nmax);
		// cout << npstr << endl;
		for (int i=0; i<(*npstr).length(); i++)
			printf("%c ",(*npstr)[i]);
		printf("\n");
	}
#endif

	//*** release memory
	printf("Releasing rect-matrices\n");
	for (int i=0; i<nthr; i++)
		cvReleaseMat(&(npcharrect[i]));

	return nmax;

}

int ilNPCharReadAtThresh(const CvArr *srcarr, string *npstr, CvMat **npcharrect, int graythresh){
	
    CvMat *src=(CvMat*)srcarr;

	//*** allocate memory
	CvMemStorage* rectstorage = cvCreateMemStorage(0);
    CvSeq *charrectseq=cvCreateSeq(0,sizeof(CvSeq),sizeof(CvRect),rectstorage);

	//*** segment characters
	int rectnum=0;
	ilCharLineSegment(src,charrectseq,&rectnum,graythresh);
	IplImage* srcgrimg=cvCreateImage(cvSize(src->width,src->height),8,1);
	cvConvert(src,srcgrimg);

#if CHARSEGMENT_DEBUG
	IplImage* imgdisp=cvCreateImage(cvSize(src->width,src->height),8,1);
	cvConvert(src,imgdisp);
#endif

	(*npstr)="";
	if (rectnum==0){
		*npcharrect=NULL;
	    cvReleaseMemStorage(&rectstorage);
	    cvReleaseImage(&srcgrimg);
		return 0;
	}

	//*** convert seq. of rectangles into CvMat
	*npcharrect=cvCreateMat(rectnum,4,CV_64FC1);
	CvRect *r=CV_GET_SEQ_ELEM(CvRect,charrectseq,0);
	double *rptr=(double*)cvPtr2D(*npcharrect,0,0);
	for (int i=0; (i<rectnum)&&(r); i++){
		rptr[i*4]=r->x;
		rptr[i*4+1]=r->y;
		rptr[i*4+2]=r->width;
		rptr[i*4+3]=r->height;
		r=CV_GET_SEQ_ELEM(CvRect,charrectseq,i+1);

#if CHARSEGMENT_DEBUG
		cvRectangle(imgdisp, cvPoint(r->x,r->y), cvPoint(r->x+r->width,r->y+r->height),cvScalar(255),1);
#endif
	}
  	
	//*** recognize characters
	ilCharRecognize(srcgrimg,charrectseq,npstr,rectnum,graythresh);
	//(*npstr)="12345678901234567890";

#if CHARSEGMENT_DEBUG
	printf("Read %d characters: ",rectnum);
	// cout << npstr << endl;
	for (int i=0; i<rectnum; i++)
		printf("%c ",(*npstr)[i]);
	printf("\n");
	cvNamedWindow("CharRectangles",1);
	cvShowImage("CharRectangles",imgdisp);
    cvWaitKey(0);
	cvReleaseImage(&imgdisp);
#endif

	//*** release memory
	cvReleaseMemStorage(&rectstorage);
	cvReleaseImage(&srcgrimg);

	return 1;
}


//************************************************************************
// ilCharRecognize
//   Given a gray image and a sequence of rectangles corresponding to
//   characters in the image does recognize a character in each rectagle
//   returns all the characters in dstseq.
//************************************************************************
void ilCharRecognize(CvArr *srcarr, CvSeq *charrectseq, string *npstrout, int rectnum, int threshval){
//void ilCharRecognize(CvArr *srcarr, CvSeq *charrectseq, char *npcharout, int rectnum){

	int rind=0, wchar=17, hchar=20;
    IplImage *src = (IplImage*)srcarr;
	IplImage *srcbin = cvCreateImage(cvSize(src->width,src->height),IPL_DEPTH_8U,1);
	IplImage *imgchar= cvCreateImage(cvSize(wchar,hchar),IPL_DEPTH_8U,1);
	CvMat *matchar = cvCreateMat(hchar,wchar,CV_64FC1);
	CvMat *matchartr=cvCreateMat(wchar,hchar,CV_64FC1);
	CvRect *r=CV_GET_SEQ_ELEM(CvRect,charrectseq,rind);
	
	//*** threshold image
	//int threshval=120;
    cvThreshold(src,srcbin,threshval, 255, CV_THRESH_BINARY );

	//*** create mx-data
	char ascitable[] = "0123456789";
    char nnfname[] = "C:/ilaptev/proj/visionapplications/numberplates/data/characters/neuralnetwork_digits_hnn20.mat";
    //char nnfname[] = "data/characters/neuralnetwork_digits_hnn20.mat";
	mxArray *mxascitable = mxCreateString(ascitable);
	mxArray *mxnnfname = mxCreateString(nnfname);
	mxArray *mxpixeldata = mxCreateDoubleMatrix(hchar*wchar,rectnum,mxREAL);
	mxArray *out = NULL;
	char c;

#if CHARSEGMENT_WRITEXML
    char matfnameout[] = "C:/temp/tmpcharimg.xml";
    CvFileStorage* fsw = cvOpenFileStorage(matfnameout,0,CV_STORAGE_WRITE);
#endif

	for (int i=0; (i<rectnum)&&(r); i++){
	//while (r){
  	    //cvRectangle(src, cvPoint(r->x,r->y), cvPoint(r->x+r->width,r->y+r->height),cvScalar(255,0,0,0),1);

		cvResetImageROI(imgchar);
		cvSet(imgchar,cvScalar(255));
		cvSetImageROI(srcbin,cvRect(r->x,r->y,r->width,r->height));
		int dstw=floor((double)hchar*(double)(r->width)/(double)(r->height));
		if (dstw<=0) dstw=1;

		cvSetImageROI(imgchar,cvRect(0,0,dstw,hchar));
		cvResize(srcbin,imgchar,CV_INTER_LINEAR);
		//*** TODO ^ unclear why the image is binary after resize(!)
	
		cvResetImageROI(imgchar);

		
	    //cvNamedWindow("CharSegImage2",1);
		//cvShowImage("CharSegImage2",imgchar);
		//cvWaitKey(0);

		cvConvert(imgchar,matchar);
		cvConvertScale(matchar,matchar,-1.0/255.0,1);

		cvTranspose(matchar,matchartr);
        memcpy(mxGetPr(mxpixeldata)+i*wchar*hchar, (double*)cvPtr2D(matchartr,0,0), wchar*hchar*sizeof(double));
  

#if CHARSEGMENT_WRITEXML
	    cvWrite(fsw,"matchar",matchar,cvAttrList(0,0));
#endif

		rind++;
		r=CV_GET_SEQ_ELEM(CvRect,charrectseq,rind);
	}
	//printf("\n");

    //*** NN character recognition
    //t1=time(&t2);
	mlfNncharrec(1,&out,mxpixeldata,mxnnfname,mxascitable);
	int elsize=mxGetElementSize(out); // should be 2 (!)
	*npstrout="";
	for (int i=0; (i<rectnum)&&(i<mxGetM(out)*mxGetN(out)); i++) 
		(*npstrout)+=(char)(((ushort*)mxGetData(out))[i]);
		//npcharout[i]=(char)(((ushort*)mxGetData(out))[i]);

	//printf("Read char %s\n",npcharout);
    //printf(" Time of matlab fun call; %1.3fsec.\n",difftime(time(&t2),t1));


#if CHARSEGMENT_WRITEXML
    cvReleaseFileStorage( &fsw );  
#endif

	mxDestroyArray(out); out=0;
	mxDestroyArray(mxpixeldata); mxpixeldata=0;
	mxDestroyArray(mxnnfname); mxnnfname=0;
	mxDestroyArray(mxascitable); mxascitable=0;

	cvReleaseImage(&srcbin);
	cvReleaseImage(&imgchar);
	cvReleaseMat(&matchar);
	cvReleaseMat(&matchartr);
}

//************************************************************************
// ilCharLineSegment
//   Finds boundindg boxes of characters in the input one text-line image
//   The segmentation is done in the pipeline: thresholding ->
//   connected components -> filtering of inliers with the constraint
//   that selected (charactr) regions should be of approximately the same
//   height and should be on the same line. RANSAC-like procedure is
//   employed for filtering. Black-on-white and white-on-black character
//   segmentation is supported.
//************************************************************************
void ilCharLineSegment(const CvArr *srcarr, CvSeq *dstseq, int *rectnumout, int graythreshval){
	
    CvMat *src=(CvMat*)srcarr;
    CvMat* srcbin = cvCreateMat(src->height,src->width,CV_8UC1);
    cvThreshold(src,srcbin,graythreshval, 255, CV_THRESH_BINARY );
	cvConvertScale(srcbin,srcbin,-1,1);

	CvMemStorage* storage = cvCreateMemStorage(0);
    CvSeq *contour = 0, *contour0;
    cvFindContours(srcbin, storage, &contour, sizeof(CvContour), CV_RETR_CCOMP, CV_CHAIN_APPROX_SIMPLE );
	contour0=contour;

	
#if CHARSEGMENT_DEBUG
	//*** display segmented regions
	CvMat *imgdisp=cvCreateMat(src->height,src->width,CV_8UC3);
	//cvCopy(src,imgdisp);
	cvCvtColor(src,imgdisp,CV_GRAY2RGB);
	
    for( ; contour != 0; contour = contour->h_next ){
        CvScalar color = CV_RGB( rand()&255, rand()&255, rand()&255 );
        /* replace CV_FILLED with 1 to see the outlines */
        cvDrawContours( imgdisp, contour, color, color, -1, CV_FILLED, 8 );
    }

	cvNamedWindow("Segmentation",1);
	cvShowImage("Segmentation",imgdisp);
	cvReleaseMat(&imgdisp);
#endif

	contour=contour0;
	CvRect rect;
	int cnum=0, cnumacc=0, i=0;
	int cnummax=3000;
	int *flagarr = new int[cnummax];
	double *rectarr = new double[4*cnummax];
	*rectnumout=0;

	for( ; contour != 0; contour = contour->h_next ){
		rect=cvBoundingRect(contour,0);
		//cvSeqPush(dstseq,&rect);
        rectarr[4*i]=rect.x;
        rectarr[4*i+1]=rect.y;
        rectarr[4*i+2]=rect.width;
        rectarr[4*i+3]=rect.height;
        flagarr[i]=1;
		cnum=i++;
    }

	if (cnum>0){
		cnumacc=ilCharLineRANSAC(rectarr, flagarr, cnum, 20);
		//cnumacc=ilCharLineRANSACUpdate1(rectarr, flagarr, cnum, 1000);
    
		//*** fill result with accepted rectangles
		for (int i=0; i<cnum; i++){
			if (flagarr[i]){
				rect.x=rectarr[4*i];
				rect.y=rectarr[4*i+1];
				rect.width=rectarr[4*i+2];
				rect.height=rectarr[4*i+3];
				cvSeqPush(dstseq,&rect);
				(*rectnumout)++;
			}
		}

		//*** sort rectangles by x coordinate
		cvSeqSort(dstseq, ilRectCmpX, 0);
	}

	delete[] flagarr;
	delete[] rectarr;

	cvReleaseMat(&srcbin);
	cvReleaseMemStorage(&storage);
}

int ilCharLineRANSAC(double *rectarr, int *flagacc, int cnum, int inum){

	if (cnum<=0)
		return 0;
	if (cnum==1){
		flagacc[0]=1;
		return 1;
	}
		
	int cnumacc=0, cnummax=0, i1=0, i2=0;
	int *flagarr = new int[cnum];
    double hm,nx,ny,norm,offset;
	for (int i=0; i<inum; i++){
		//*** take two random rectangles
		i1=0; i2=0;
		while (i1==i2){
			i1=floor(((double)cnum+1)*((double)rand())/32767.0);
			i2=floor(((double)cnum+1)*((double)rand())/32767.0);
		}
		//*** estimate line normal and mean height
		hm=(rectarr[4*i1+3]+rectarr[4*i2+3])/2.0;
		nx=rectarr[4*i1+1]-rectarr[4*i2+1];
		ny=-(rectarr[4*i1]-rectarr[4*i2]);
		norm=sqrt(nx*nx+ny*ny);
		nx=nx/norm; ny=ny/norm;
		offset=nx*rectarr[4*i1]+ny*rectarr[4*i1+1];

		//*** find char-boxes on the line
		cnumacc=0;
		for (int j=0; j<cnum; j++){
			if ((abs(log(rectarr[4*j+3]/hm))<0.3) &&						// height constraint
				(abs(nx*rectarr[4*j]+ny*rectarr[4*j+1]-offset)/hm<0.2) &&	// distance from the line constraint
				(rectarr[4*j+2]<hm)){										// width constraint
				flagarr[j]=1;
				cnumacc++;
			} else
				flagarr[j]=0;
		}
		
		//*** update best estimate
		if (cnumacc>cnummax){
			cnummax=cnumacc;
			for (int j=0; j<cnum; j++)
				flagacc[j]=flagarr[j];
		}
	}	

    delete[] flagarr;
	return cnummax;
}

int ilCharLineRANSACUpdate1(double *rectarr, int *flagacc, int cnum, int inum){

	if (cnum<=0)
		return 0;
	if (cnum==1){
		flagacc[0]=1;
		return 1;
	}
		
	int cnumacc=0, cnummax=0, i1=0, i2=0;
	int *flagarr = new int[cnum];
    double hm,nx,ny,norm,offset;
	int x1,y1,h1,w1,x2,y2,h2,w2;
	for (int i=0; i<inum; i++){
		//*** take two random rectangles
		i1=0; i2=0;
		while (i1==i2){
			i1=floor(((double)cnum+1)*((double)rand())/32767.0);
			i2=floor(((double)cnum+1)*((double)rand())/32767.0);
		}
		x1=rectarr[4*i1];   x2=rectarr[4*i2];
		y1=rectarr[4*i1+1]; y2=rectarr[4*i2+1];
		w1=rectarr[4*i1+2]; w2=rectarr[4*i2+2];
		h1=rectarr[4*i1+3]; h2=rectarr[4*i2+3];
		if ((abs(y1-y2)<h1)&&(abs(y1-y2)<h2)&&
			(abs(x1-x2)<5*w1)&&(abs(x1-x2)<5*w2)&&
			(w1/w2<3)&&(w2/w1<3)&&(h1/h2<2)&&(h2/h1<2)){

			//*** estimate line normal and mean height

			hm=(rectarr[4*i1+3]+rectarr[4*i2+3])/2.0;
			nx=rectarr[4*i1+1]-rectarr[4*i2+1];
			ny=-(rectarr[4*i1]-rectarr[4*i2]);
			norm=sqrt(nx*nx+ny*ny);
			nx=nx/norm; ny=ny/norm;
			offset=nx*rectarr[4*i1]+ny*rectarr[4*i1+1];

			//*** find char-boxes on the line
			cnumacc=0;
			for (int j=0; j<cnum; j++){
				if ((abs(log(rectarr[4*j+3]/hm))<0.3) &&						// height constraint
					(abs(nx*rectarr[4*j]+ny*rectarr[4*j+1]-offset)/hm<0.2) &&	// distance from the line constraint
					(rectarr[4*j+2]<hm)){										// width constraint
					flagarr[j]=1;
					cnumacc++;
				} else
					flagarr[j]=0;
			}
		
			//*** update best estimate
			if (cnumacc>cnummax){
				cnummax=cnumacc;
				for (int j=0; j<cnum; j++)
					flagacc[j]=flagarr[j];
			}
		}
	}

    delete[] flagarr;
	return cnummax;
}

static int ilRectCmpX( const void* _a, const void* _b, void* userdata )
{
    CvRect* a = (CvRect*)_a;
    CvRect* b = (CvRect*)_b;
	return a->x-b->x;
}
