#include "stdafx.h"
#include "integralhistogram.h"
#include "ilmatrixoperations.h"

#define TEXTONCLASSIFY_WRITEXMLRESULTS 0

//*********************************************************************
// ilIntegralBlockHist
//   Computes integral weighted block histogram from an image with labeled pixels
//   Inputs timg: image with labeled pixels; wimg: image with pixel weights;
//          nc: number of labels; bxsz,bysz: dimensions of the block
//*********************************************************************
void ilIntegralBlockHist(const CvArr* timgarr, const CvArr* wimgarr, CvMat** ihist, int nc, double bxsz, double bysz)
{	
    CvMat *timg=(CvMat*)timgarr;
    CvMat *wimg=(CvMat*)wimgarr;
    int ysz=timg->rows, xsz=timg->cols, pixind;
	int nby=ihist[0]->rows, nbx=ihist[0]->cols;

	//*** pre-compute block indices of pixels along x and y directions
	int *bxind = new int[xsz];
	int *byind = new int[ysz];
	for (int i=0; i<xsz; i++)
		bxind[i]=min(nbx-1,floor((double)i/bxsz));
	for (int i=0; i<ysz; i++)
		byind[i]=min(nby-1,floor((double)i/bysz));
 
	//*** set up pointers
	short *ptrtimg=(short*)cvPtr2D(timg,0,0);
    float *ptrwimg;
	if (wimg) ptrwimg=(float*)cvPtr2D(wimg,0,0);
	float **ptrihist = new float*[nc];
	for (int i=0; i<nc; i++) ptrihist[i]=(float*)cvPtr2D(ihist[i],0,0);

	//*** initialize histogram values to zero
	for (int i=0; i<nc; i++)
		for (int j=0; j<nbx*nby; j++)
			ptrihist[i][j]=0.0;

	//*** loop over image pixels and accumulate block histograms
	for (int i=0; i<ysz; i++){
		for (int j=0; j<xsz; j++){
			int bind=byind[i]*nbx+bxind[j]; // block index
			int pind=i*xsz+j;               // pixel index
			int k=ptrtimg[pind];            // hist-bin index
			if (!wimg)
				ptrihist[k][bind]+=1.0;
			else
				ptrihist[k][bind]+=ptrwimg[pind];
		}
	}			

	//*** integrate histogram
	for (int i=0; i<nby; i++){
		for (int j=1; j<nbx; j++){
			int bind=i*nbx+j;
			for (int k=0; k<nc; k++) {
				ptrihist[k][bind]+=ptrihist[k][bind-1];
			}
		}
	}
	for (int i=1; i<nby; i++){
		for (int j=0; j<nbx; j++){
			int bind=i*nbx+j;
			for (int k=0; k<nc; k++) {
				ptrihist[k][bind]+=ptrihist[k][bind-nbx];
			}
		}
	}

	delete[] bxind;
	delete[] byind;
	delete[] ptrihist;

#if TEXTONCLASSIFY_WRITEXMLRESULTS
    char matfnameout[] = "C:/temp/tmpinthist.xml";
    CvFileStorage* fsw = cvOpenFileStorage(matfnameout,0,CV_STORAGE_WRITE);
	cvWrite(fsw,"timg",timg,cvAttrList(0,0));
	for (int i=0; i<nc; i++) cvWrite(fsw,"blockhist",ihist[i],cvAttrList(0,0));
    cvReleaseFileStorage( &fsw );  
#endif
}

//*********************************************************************
// ilIntegralBlockHistOrig
//   Computes integral weighted block histogram from an image with labeled pixels
//   Inputs timg: image with labeled pixels; wimg: image with pixel weights;
//          nc: number of labels; bxsz,bysz: dimensions of the block
//*********************************************************************
void ilIntegralBlockHistOrig(const CvArr* timgarr, const CvArr* wimgarr, CvMat** ihist, int nc, int bxsz, int bysz)
{	
    CvMat *timg=(CvMat*)timgarr;
    CvMat *wimg=(CvMat*)wimgarr;
    int ysz=timg->rows, xsz=timg->cols, pixind;
	int nby=ihist[0]->rows, nbx=ihist[0]->cols;

	// pre-compute indices of pixels in a block
	int *pixindinblock = new int[bxsz*bysz];
	//for (int i=0; i<bxsz; i++)
	//	for (int j=0; j<bysz; j++)
	//		pixindinblock[i*bysz+j]=i*xsz+j;
	for (int i=0; i<bysz; i++)
		for (int j=0; j<bxsz; j++)
			pixindinblock[i*bxsz+j]=i*xsz+j;

	// set up pointers
	short *ptrtimg=(short*)cvPtr2D(timg,0,0);
    float *ptrwimg;
	if (wimg) ptrwimg=(float*)cvPtr2D(wimg,0,0);
	float **ptrihist = new float*[nc];
	for (int i=0; i<nc; i++) ptrihist[i]=(float*)cvPtr2D(ihist[i],0,0);

	//*** compute block histogram
	float *hblock = new float[nc];
	for (int i=0; i<nby; i++){
		for (int j=0; j<nbx; j++){
			int baseind=i*bxsz*xsz+j*bysz;
		    for (int k=0; k<nc; k++) hblock[k]=0.0;
			for (int k=0; k<bxsz*bysz; k++){
				if (!wimg) {
					hblock[ptrtimg[baseind+pixindinblock[k]]]+=1.0;
				} else {
					pixind=baseind+pixindinblock[k];
					hblock[ptrtimg[pixind]]+=ptrwimg[pixind];
				}
			}
			int bind=i*nbx+j;
			for (int k=0; k<nc; k++) {
				ptrihist[k][bind]=hblock[k];
			}
		}
	}
	//*** integrate histogram
	for (int i=0; i<nby; i++){
		for (int j=1; j<nbx; j++){
			int bind=i*nbx+j;
			for (int k=0; k<nc; k++) {
				ptrihist[k][bind]+=ptrihist[k][bind-1];
			}
		}
	}
	for (int i=1; i<nby; i++){
		for (int j=0; j<nbx; j++){
			int bind=i*nbx+j;
			for (int k=0; k<nc; k++) {
				ptrihist[k][bind]+=ptrihist[k][bind-nbx];
			}
		}
	}

	delete[] pixindinblock;
	delete[] ptrihist;
	delete[] hblock;

#if TEXTONCLASSIFY_WRITEXMLRESULTS
    char matfnameout[] = "C:/temp/tmpinthist.xml";
    CvFileStorage* fsw = cvOpenFileStorage(matfnameout,0,CV_STORAGE_WRITE);
	cvWrite(fsw,"timg",timg,cvAttrList(0,0));
	for (int i=0; i<nc; i++) cvWrite(fsw,"blockhist",ihist[i],cvAttrList(0,0));
    cvReleaseFileStorage( &fsw );  
#endif
}

//*********************************************************************
// ilClassifyImagePixels
//   Computes labels for all image pixels by (a) applying local filters
//   at each location and (b) projecting resulting features onto pre-defined
//   directions (could be textons or orientation directions) in a feature
//   space. The absolute value and the label of maximized projections
//   are returned.
//
//*********************************************************************
void ilClassifyImagePixels(const CvArr* srcarr, CvArr* timgarr, CvArr* wimgarr, CvArr* cmeansnormarr, int pixclasstype){	
    double dt;
    int t1,t2;
    CvMat *src=(CvMat*)srcarr;
    CvMat *timg=(CvMat*)timgarr;
    CvMat *wimg=(CvMat*)wimgarr;
    CvMat *cmeansnorm=(CvMat*)cmeansnormarr;
	int ysz=src->rows, xsz=src->cols; //, nc=cmeansnorm->rows;

    t1=cvGetTickCount();
	if (pixclasstype==TEXTONS_SC12_GDORDER2_)
 		ilTextonClassifySC12GDO2(src,timg,wimg,cmeansnorm);
	if (pixclasstype==ORIENTGRAD_DNUM4_)
		ilOrientGradClassifyD4(src,timg,wimg);
    t2=cvGetTickCount();
    dt=(double(t2-t1)/cvGetTickFrequency())/pow(10.0,6);

#if TEXTONCLASSIFY_WRITEXMLRESULTS
    char matfnameout[] = "C:/temp/tmptextons.xml";
    CvFileStorage* fsw = cvOpenFileStorage(matfnameout,0,CV_STORAGE_WRITE);
    cvWrite(fsw,"textons",timg,cvAttrList(0,0));
    cvWrite(fsw,"weights",wimg,cvAttrList(0,0));
    cvReleaseFileStorage( &fsw );  
#endif
}

//*********************************************************************
// ilTextonClassify
//   Computes texton labels for all image points.
//
//*********************************************************************
void ilTextonClassifySC12GDO2(const CvArr* srcarr, CvArr* textonsarr, CvArr* weightsarr, CvArr *cmeansnormarr)
{
  CvMat *src=(CvMat*)srcarr;
  CvMat *textons=(CvMat*)textonsarr;
  CvMat *weights=(CvMat*)weightsarr;
  CvMat *cmeansnorm=(CvMat*)cmeansnormarr;
  int ysz=src->rows, xsz=src->cols, cnum=cmeansnorm->rows, dnum=cmeansnorm->cols;
  CvMat *src32f = cvCreateMat(ysz,xsz,CV_32FC1);
  cvConvert(src,src32f);

  //*** allocate space for image filtering
  CvMat* data = cvCreateMat(dnum,ysz*xsz,CV_32FC1);
  CvMat* imagegraysmooth[2];
  CvMat* imagegrayderiv[10];

  for (int i=0; i<2; i++){
	imagegraysmooth[i] = cvCreateMat(ysz,xsz,CV_32FC1);
  }

  int fsize=sizeof(float);
  for (int i=0; i<10; i++){
	imagegrayderiv[i] = cvCreateMatHeader(ysz,xsz,CV_32FC1);
	cvSetData(imagegrayderiv[i],cvPtr2D(data,i,0),xsz*sizeof(float));
  }
  
  int colbytelen=cvPtr2D(data,1,0)-cvPtr2D(data,0,0);

  //*** compute Gaussian derivatives
  double dt;
  int t1,t2;
  t1=cvGetTickCount();
  cvSmooth(src32f, imagegraysmooth[0], CV_GAUSSIAN, 0, 0, 1.0);
  cvSmooth(src32f, imagegraysmooth[1], CV_GAUSSIAN, 0, 0, 2.0);
  cvSobel(imagegraysmooth[0],imagegrayderiv[0],int(1),int(0),int(1));
  cvSobel(imagegraysmooth[0],imagegrayderiv[2],int(0),int(1),int(1));
  cvSobel(imagegraysmooth[0],imagegrayderiv[4],int(2),int(0),int(1));
  cvSobel(imagegraysmooth[0],imagegrayderiv[6],int(1),int(1),int(3));
  cvSobel(imagegraysmooth[0],imagegrayderiv[8],int(0),int(2),int(1));
  cvSobel(imagegraysmooth[1],imagegrayderiv[1],int(1),int(0),int(1));
  cvSobel(imagegraysmooth[1],imagegrayderiv[3],int(0),int(1),int(1));
  cvSobel(imagegraysmooth[1],imagegrayderiv[5],int(2),int(0),int(1));
  cvSobel(imagegraysmooth[1],imagegrayderiv[7],int(1),int(1),int(3));
  cvSobel(imagegraysmooth[1],imagegrayderiv[9],int(0),int(2),int(1));

  //*** scale 1st order partial derivatives for the correspondance with Matlab
  cvScale(imagegrayderiv[0],imagegrayderiv[0],0.5);
  cvScale(imagegrayderiv[1],imagegrayderiv[1],0.5);
  cvScale(imagegrayderiv[2],imagegrayderiv[2],0.5);
  cvScale(imagegrayderiv[3],imagegrayderiv[3],0.5);
  cvScale(imagegrayderiv[6],imagegrayderiv[6],0.25);
  cvScale(imagegrayderiv[7],imagegrayderiv[7],0.25);
  t2=cvGetTickCount();
  dt=(double(t2-t1)/cvGetTickFrequency())/pow(10.0,6);

  
  //*** Save intermidiate results to xml
#if TEXTONCLASSIFY_WRITEXMLRESULTS
  char matfnameout[] = "C:/temp/tmp.xml";
  CvFileStorage* fsw = cvOpenFileStorage(matfnameout,0,CV_STORAGE_WRITE);
  cvWrite(fsw,"fdata",data,cvAttrList(0,0));
#endif

  //*** project onto normalized clusters
  t1=cvGetTickCount();
  CvMat* cdata = cvCreateMat(cnum,ysz*xsz,CV_32FC1);
  cvGEMM(cmeansnorm,data,1.0,0,0,cdata);
  t2=cvGetTickCount();
  dt=(double(t2-t1)/cvGetTickFrequency())/pow(10.0,6);

  //*** maximize projection over clusters
  t1=cvGetTickCount();
  CvMat *maxind = cvCreateMatHeader(ysz*xsz,1,CV_16SC1);
  CvMat* maxval = cvCreateMatHeader(ysz*xsz,1,CV_32FC1);
  cvSetData(maxind,cvPtr2D(textons,0,0),sizeof(short));
  cvSetData(maxval,cvPtr2D(weights,0,0),sizeof(float));
  ilMaxColFloat(cdata,maxval,maxind);
  //ilMaxColFloat(cdata,maxval,maxind,1); // with abs-flag ON ??
  t2=cvGetTickCount();
  dt=(double(t2-t1)/cvGetTickFrequency())/pow(10.0,6);

#if TEXTONCLASSIFY_WRITEXMLRESULTS
  cvWrite(fsw,"cdata",cdata,cvAttrList(0,0));
  cvWrite(fsw,"clusterind",maxind,cvAttrList(0,0));
  cvWrite(fsw,"clusterdst",maxval,cvAttrList(0,0));
  cvWrite(fsw,"cmeansnormarr",cmeansnormarr,cvAttrList(0,0));
#endif

#if TEXTONCLASSIFY_WRITEXMLRESULTS
  cvReleaseFileStorage( &fsw );  
#endif

  //*** release space
  t1=cvGetTickCount();
  cvReleaseMat(&data);
  cvReleaseMat(&cdata);
  
  //cvReleaseMat(&src32f); //-- ???
  for (int i=0; i<2; i++) cvReleaseMat(&imagegraysmooth[i]);
  t2=cvGetTickCount();
  dt=(double(t2-t1)/cvGetTickFrequency())/pow(10.0,6); 
 }

//*********************************************************************
// ilOrientGradClassify
//   Computes discretized gradient orientation at each point.
//
//*********************************************************************
void ilOrientGradClassifyD4(const CvArr* srcarr, CvArr* textonsarr, CvArr* weightsarr)
{
  CvMat *src=(CvMat*)srcarr;
  CvMat *textons=(CvMat*)textonsarr;
  CvMat *weights=(CvMat*)weightsarr;
  int ndirs=4;
  CvMat *cmeansnorm=cvCreateMat(ndirs,2,CV_32FC1);
  float *ptrcmeansnorm=(float*)cvPtr2D(cmeansnorm,0,0);
  for (int i=0; i<ndirs; i++){
	ptrcmeansnorm[i*2]=float(cos(float(i)*3.14159265358979323846/float(ndirs)));
	ptrcmeansnorm[i*2+1]=float(sin(float(i)*3.14159265358979323846/float(ndirs)));
  }

  int ysz=src->rows, xsz=src->cols, cnum=cmeansnorm->rows, dnum=cmeansnorm->cols;
  CvMat *src32f = cvCreateMat(ysz,xsz,CV_32FC1);
  cvConvert(src,src32f);

  //*** allocate space for image filtering
  CvMat* data = cvCreateMat(dnum,ysz*xsz,CV_32FC1);
  CvMat* imagegraysmooth=cvCreateMat(ysz,xsz,CV_32FC1);
  CvMat* imagegrayderiv[2];

  int fsize=sizeof(float);
  for (int i=0; i<2; i++){
	imagegrayderiv[i] = cvCreateMatHeader(ysz,xsz,CV_32FC1);
	cvSetData(imagegrayderiv[i],cvPtr2D(data,i,0),xsz*sizeof(float));
  }
  
  int colbytelen=cvPtr2D(data,1,0)-cvPtr2D(data,0,0);

  //*** compute Gaussian derivatives
  double dt;
  int t1,t2;
  t1=cvGetTickCount();
  cvSmooth(src32f, imagegraysmooth, CV_GAUSSIAN, 0, 0, 1.0);
  cvSobel(imagegraysmooth,imagegrayderiv[0],int(1),int(0),int(1));
  cvSobel(imagegraysmooth,imagegrayderiv[1],int(0),int(1),int(1));

  //*** scale 1st order partial derivatives for the correspondance with Matlab
  cvScale(imagegrayderiv[0],imagegrayderiv[0],0.5);
  cvScale(imagegrayderiv[1],imagegrayderiv[1],0.5);
  t2=cvGetTickCount();
  dt=(double(t2-t1)/cvGetTickFrequency())/pow(10.0,6);

  //*** Save intermidiate results to xml
#if TEXTONCLASSIFY_WRITEXMLRESULTS
  char matfnameout[] = "C:/temp/tmp.xml";
  CvFileStorage* fsw = cvOpenFileStorage(matfnameout,0,CV_STORAGE_WRITE);
  cvWrite(fsw,"fdata",data,cvAttrList(0,0));
#endif

  //*** project onto normalized clusters
  t1=cvGetTickCount();
  CvMat* cdata = cvCreateMat(cnum,ysz*xsz,CV_32FC1);
 
  //*** this should work when OpenCV fixes the bug with cvGEMM
  //cvGEMM(cmeansnorm,data,1.0,0,1,cdata);
  //*** this is a specialized substitute for cvGEMM
  ilOrientGradClassifyD4Project(cmeansnorm,data,cdata);

  t2=cvGetTickCount();
  dt=(double(t2-t1)/cvGetTickFrequency())/pow(10.0,6);

  //*** maximize projection over clusters
  t1=cvGetTickCount();
  CvMat *maxind = cvCreateMatHeader(ysz*xsz,1,CV_16SC1);
  CvMat *maxval = cvCreateMatHeader(ysz*xsz,1,CV_32FC1);
  cvSetData(maxind,cvPtr2D(textons,0,0),sizeof(short));
  cvSetData(maxval,cvPtr2D(weights,0,0),sizeof(float));
  ilMaxColFloat(cdata,maxval,maxind,1); // with abs-flag ON
  t2=cvGetTickCount();
  dt=(double(t2-t1)/cvGetTickFrequency())/pow(10.0,6);

#if TEXTONCLASSIFY_WRITEXMLRESULTS
  cvWrite(fsw,"cdata",cdata,cvAttrList(0,0));
#endif

#if TEXTONCLASSIFY_WRITEXMLRESULTS
  cvWrite(fsw,"clusterind",maxind,cvAttrList(0,0));
  cvWrite(fsw,"clusterdst",maxval,cvAttrList(0,0));
#endif

#if TEXTONCLASSIFY_WRITEXMLRESULTS
  cvReleaseFileStorage( &fsw );  
#endif

  //*** release space
  t1=cvGetTickCount();
  cvReleaseMat(&data);
  cvReleaseMat(&cdata);
  cvReleaseMat(&imagegraysmooth);
  cvReleaseMat(&src32f);
  cvReleaseMat(&cmeansnorm);
  
  cvReleaseMatHeader(&maxind);
  cvReleaseMatHeader(&maxval);
  for (int i=0; i<2; i++) cvReleaseMatHeader(&imagegrayderiv[i]);

  t2=cvGetTickCount();
  dt=(double(t2-t1)/cvGetTickFrequency())/pow(10.0,6); 
 }

void ilOrientGradClassifyD4Project(const CvArr* srcarr1, const CvArr* srcarr2, CvArr* dstarr)
{
  CvMat *src1=(CvMat*)srcarr1;
  CvMat *src2=(CvMat*)srcarr2;
  CvMat *dst =(CvMat*)dstarr;
  float *ptrsrc1=(float*)cvPtr2D(src1,0,0);
  float *ptrsrc2=(float*)cvPtr2D(src2,0,0);
  float *ptrdst=(float*)cvPtr2D(dst,0,0);
  long n=src2->cols;

  //*** assume src1 is a 4x2 float matrix
  float m11=ptrsrc1[0]; float m12=ptrsrc1[1];
  float m21=ptrsrc1[2]; float m22=ptrsrc1[3];
  float m31=ptrsrc1[4]; float m32=ptrsrc1[5];
  float m41=ptrsrc1[6]; float m42=ptrsrc1[7];
  float v1,v2;

  for (long i=0; i<n; i++){
    v1=ptrsrc2[i]; v2=ptrsrc2[i+n];
    ptrdst[i]    =m11*v1+m12*v2;
    ptrdst[i+n]  =m21*v1+m22*v2;
    ptrdst[i+2*n]=m31*v1+m32*v2;
    ptrdst[i+3*n]=m41*v1+m42*v2;
  }
}